/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#define __STDC_LIMIT_MACROS
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <cstring>
#include <exception>
#include "ARINC424_18.h"
#include "Airport.h"
#include "Util.h"

Airport::Airport(std::string IATA,std::string ICAO)
{
	this->IATA=IATA;
	this->ICAO=ICAO;
	LocType=Waypoint::LOC_AIRPORT;
	MagVar=NAN;
	Elevation=0;
}

Airport::~Airport() {
	for(std::vector<Runway*>::iterator Runwayi = Runways.begin(); Runwayi != Runways.end(); ++Runwayi)
	{
		delete *Runwayi;
	}
	Runways.clear();

	for(std::vector<NavAid*>::iterator NavAidi = NavAids.begin(); NavAidi != NavAids.end(); ++NavAidi)
	{
		delete *NavAidi;
	}
	NavAids.clear();

	for(std::vector<Frequency*>::iterator Freqi = Freqs.begin(); Freqi != Freqs.end(); ++Freqi)
	{
		delete *Freqi;
	}
	Freqs.clear();

	ClearApproachPlates();
}

std::string Airport::GetIATA()
{
	return IATA;
}

std::string Airport::GetICAO()
{
	return ICAO;
}

double Airport::GetMagVar()
{
	return MagVar;
}

int Airport::GetElevation()
{
	return Elevation;
}

std::vector<Frequency*> *Airport::GetFreqs()
{
	return &Freqs;
}

std::vector<Runway*> *Airport::GetRunways()
{
	return &Runways;
}

std::vector<NavAid*> *Airport::GetNavAids()
{
	return &NavAids;
}

std::vector<ApproachPlate*> *Airport::GetApproachPlates()
{
	return &ApproachPlates;
}

std::vector<METAR*> *Airport::GetMETARs()
{
	return &METARs;
}

void Airport::AddRunway(Runway *R)
{
	Runways.push_back(R);
}

void Airport::AddNavAid(NavAid *N)
{
	NavAids.push_back(N);
}

void Airport::AddFrequency(Frequency *F)
{
	Freqs.push_back(F);
}

void Airport::AddApproachPlate(ApproachPlate *A)
{
	ApproachPlates.push_back(A);
}

void Airport::ClearApproachPlates()
{
	for(std::vector<ApproachPlate*>::iterator ApproachPlatei = ApproachPlates.begin(); ApproachPlatei != ApproachPlates.end(); ++ApproachPlatei)
	{
		delete *ApproachPlatei;
	}
	ApproachPlates.clear();
}

void Airport::AddMETAR(METAR *M)
{
	METAR *LocalMETAR = new METAR(*M);
	METARs.push_back(LocalMETAR);
}

void Airport::ClearMETARs()
{
	for(std::vector<METAR*>::iterator METARi = METARs.begin(); METARi != METARs.end(); ++METARi)
	{
		delete *METARi;
	}
	METARs.clear();
}

bool Airport::ParseNFD(std::string Line)
{
	struct arinc424_18_airport_struct *NFDData;
	char strbuf[sizeof(struct arinc424_18_airport_struct)];

	if (Line.length()!=sizeof(struct arinc424_18_airport_struct)+1)	/* Allow for LF */
	{
		return false;
	}

	if (strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_AIRPORT)!=0)
	{
		return false;
	}

	NFDData=(struct arinc424_18_airport_struct *)Line.c_str();

	/* IATA */
	memcpy(strbuf,NFDData->header.ataiata,sizeof(NFDData->header.ataiata));
	strbuf[sizeof(NFDData->header.ataiata)]=0;
	/* Trim trailing spaces */
	while (isspace(strbuf[strlen(strbuf)-1]) && strlen(strbuf))
		strbuf[strlen(strbuf)-1]=0;
	IATA=strbuf;

	/* ICAO & ID */
	memcpy(strbuf,NFDData->header.icaoid,sizeof(NFDData->header.icaoid));
	strbuf[sizeof(NFDData->header.icaoid)]=0;
	/* Trim trailing spaces */
	while (isspace(strbuf[strlen(strbuf)-1]) && strlen(strbuf))
		strbuf[strlen(strbuf)-1]=0;
	ICAO=strbuf;
	ID=strbuf;

	/* Latitude */
	Lat=arinc424_18_parse_lat(NFDData->lat);
	if (UTIL_isnan(Lat))
		return false;

	/* Longitude */
	Lon=arinc424_18_parse_lon(NFDData->lon);
	if (UTIL_isnan(Lon))
		return false;

	/* MagVar */
	memcpy(strbuf,NFDData->magvar,sizeof(NFDData->magvar));
	strbuf[sizeof(NFDData->magvar)]=0;
	MagVar=atoi(&(strbuf[1]))/10.0;
	if (strbuf[0]=='W') {
		MagVar*=-1.0;
	}

	/* Elevation */
	memcpy(strbuf,NFDData->elevation,sizeof(NFDData->elevation));
	strbuf[sizeof(NFDData->elevation)]=0;
	Elevation=atoi(strbuf);

	/* Name */
	memcpy(strbuf,NFDData->name,sizeof(NFDData->name));
	strbuf[sizeof(NFDData->name)]=0;
	/* Trim trailing spaces */
	while (isspace(strbuf[strlen(strbuf)-1]) && strlen(strbuf))
		strbuf[strlen(strbuf)-1]=0;
	Name=strbuf;

	return true;
}

bool Airport::ParseXPlane850(std::string *Line) {
	char *tok;
	char *buf;

	/* Copy to a normal C buffer for strtok_r() to work with */
	buf=(char*)malloc(Line->length()+1);
	if (!buf)
		return false;
	strncpy(buf,Line->c_str(),Line->length());
	buf[Line->length()]=0;

	try {
		/* Parse the fields */
		if (!(tok=strtok(buf," \t")))
			throw std::exception();
		/* Ignore the row code */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Elevation=atoi(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore the tower flag */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore the deprecated data */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		ID=tok;
		ICAO=tok;
		IATA=&(tok[1]);	/* Skip the first char */
		if (!(tok=strtok(NULL,"")))	/* The remainder */
			throw std::exception();
		Name=tok;
	}

	catch (...) {
		free(buf);
		return false;
	}

	MagVar=0.0;	/* Unspecified */

	free(buf);
	return true;
}

void Airport::ComputeLocByRunways() {
	Lon=0.0;
	Lat=0.0;

	if (Runways.size()==0) {
		return;
	}

	for (std::vector<Runway*>::iterator Runwayi=Runways.begin();Runwayi!=Runways.end();Runwayi++) {
		Lon+=(*Runwayi)->GetLon();
		Lat+=(*Runwayi)->GetLat();
	}
	Lon/=Runways.size();
	Lat/=Runways.size();
}

bool Airport::MakeBinary(struct Airport_Binary_struct *BinaryData) {
	double Lonx1000000,Latx1000000, MagVarx100;

	Lonx1000000=Lon*1000000.0;
	Latx1000000=Lat*1000000.0;
	MagVarx100=MagVar*100;

	/* Sanity checking */
	if (
		ID.size()>sizeof(BinaryData->ID) ||
		Name.size()>sizeof(BinaryData->Name) ||
		IATA.size()>sizeof(BinaryData->IATA) ||
		ICAO.size()>sizeof(BinaryData->ICAO) ||
		Lonx1000000<INT32_MIN || Lonx1000000>INT32_MAX ||
		Latx1000000<INT32_MIN || Latx1000000>INT32_MAX ||
		MagVarx100>INT16_MAX || MagVarx100<INT16_MIN ||
		Elevation>INT16_MAX || Elevation<INT16_MIN ||
		Runways.size()>UINT8_MAX ||
		NavAids.size()>UINT8_MAX ||
		Freqs.size()>UINT8_MAX ||
		ApproachPlates.size()>UINT8_MAX
		) {
		return false;
	}

	strncpy(BinaryData->ID,ID.c_str(),sizeof(BinaryData->ID));
	strncpy(BinaryData->Name,Name.c_str(),sizeof(BinaryData->Name));
	strncpy(BinaryData->IATA,IATA.c_str(),sizeof(BinaryData->IATA));
	strncpy(BinaryData->ICAO,ICAO.c_str(),sizeof(BinaryData->ICAO));
	BinaryData->Lonx1000000=(int32_t)round(Lonx1000000);
	BinaryData->Latx1000000=(int32_t)round(Latx1000000);
	BinaryData->MagVarx100=(int32_t)round(MagVarx100);
	BinaryData->Elevation=Elevation;
	BinaryData->nRunways=Runways.size();
	BinaryData->nNavAids=NavAids.size();
	BinaryData->nFreqs=Freqs.size();
	BinaryData->nApproachPlates=ApproachPlates.size();

	return true;
}

void Airport::SetFromBinary(struct Airport_Binary_struct *BinaryData) {
	ID.assign(BinaryData->ID,sizeof(BinaryData->ID));
	Name.assign(BinaryData->Name,sizeof(BinaryData->Name));
	IATA.assign(BinaryData->IATA,sizeof(BinaryData->IATA));
	ICAO.assign(BinaryData->ICAO,sizeof(BinaryData->ICAO));
	/* Reset the size to the actual number of characters */
	ID=ID.c_str();
	Name=Name.c_str();
	IATA=IATA.c_str();
	ICAO=ICAO.c_str();
	Lon=BinaryData->Lonx1000000/1000000.0;
	Lat=BinaryData->Latx1000000/1000000.0;
	MagVar=BinaryData->MagVarx100/100.0;
	Elevation=BinaryData->Elevation;
}

bool Airport::Matchi(std::string str) {
	str=Util::toUpper(str);

	if (
		(Util::toUpper(IATA).substr(0,str.length())==str) ||
		(Util::toUpper(ICAO).substr(0,str.length())==str) ||
		(Util::toUpper(Name).find(str)!=std::string::npos)

	)
	{
		return true;
	}
	else
	{
		return false;
	}
}

/* The STL sort functions operation on objects, not pointers to objects.  Since
 * it's common to have a vector of pointers to airports we have this helper
 * to handle the indirection and do the sort.
 */
bool Airport::IndirectLessThan (Airport* left,Airport *right) {
	if (left && right)
		return left->GetICAO().compare(right->GetICAO())<0;
	else
		return false;
}

bool operator < (const Airport& left, const Airport& right)
{
	return ((Airport)left).GetICAO().compare(((Airport)right).GetICAO())<0;
}

bool operator == (const Airport& left, const Airport& right)
{
	return ((Airport)left).GetICAO().compare(((Airport)right).GetICAO())==0;
}
