/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <cstring>
#include <cmath>
#include "EnrouteWaypoint.h"
#include "ARINC424_18.h"
#include "Util.h"

EnrouteWaypoint::EnrouteWaypoint()
{
	LocType=Waypoint::LOC_ENROUTEWAYPOINT;
}

bool EnrouteWaypoint::ParseNFD(std::string Line)
{
	struct arinc424_18_waypoint_struct *NFDData;
	char strbuf[sizeof(struct arinc424_18_waypoint_struct)];

	if (Line.length()!=sizeof(struct arinc424_18_waypoint_struct)+1)	/* Allow for LF */
	{
		return false;
	}

	if (strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_ENROUTE_WAYPOINT)!=0)
	{
		return false;
	}

	NFDData=(struct arinc424_18_waypoint_struct *)Line.c_str();

	/* ID */
	memcpy(strbuf,NFDData->header.id,sizeof(NFDData->header.id));
	strbuf[sizeof(NFDData->header.id)]=0;
	ID=strbuf;

	/* Longitude & Latitude */
	Lat=arinc424_18_parse_lat(NFDData->lat);
	Lon=arinc424_18_parse_lon(NFDData->lon);

	if (UTIL_isnan(Lat) || UTIL_isnan(Lon))
		return false;

	/* Name */
	memcpy(strbuf,NFDData->name,sizeof(NFDData->name));
	strbuf[sizeof(NFDData->name)]=0;
	/* Trim trailing spaces */
	while (isspace(strbuf[strlen(strbuf)-1]) && strlen(strbuf))
		strbuf[strlen(strbuf)-1]=0;
	Name=strbuf;

	return true;
}

bool EnrouteWaypoint::ParseXPlane600(std::string *Line) {
	char *tok;
	char *buf;

	/* Copy to a normal C buffer for strtok_r() to work with */
	buf=(char*)malloc(Line->length()+1);
	if (!buf)
		return false;
	strncpy(buf,Line->c_str(),Line->length());
	buf[Line->length()]=0;

	try {
		/* Parse the fields */
		if (!(tok=strtok(buf," \t")))
			throw std::exception();
		Lat=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lon=atof(tok);
		if (!(tok=strtok(NULL,"")))
			throw std::exception();
		ID=tok;
		Name=tok;
	}

	catch (...) {
		free(buf);
		return false;
	}

	free(buf);
	return true;
}

bool EnrouteWaypoint::MakeBinary(struct EnrouteWaypoint_Binary_struct *BinaryData) {
	double Lonx1000000,Latx1000000;

	Lonx1000000=Lon*1000000.0;
	Latx1000000=Lat*1000000.0;
	/* Sanity checking */
	if (
		ID.size()>sizeof(BinaryData->ID) ||
		Lonx1000000<INT32_MIN || Lonx1000000>INT32_MAX ||
		Latx1000000<INT32_MIN || Latx1000000>INT32_MAX ||
		Name.size()>sizeof(BinaryData->Name)
		) {
		return false;
	}

	strncpy(BinaryData->ID,ID.c_str(),sizeof(BinaryData->ID));
	BinaryData->Lonx1000000=(int32_t)round(Lonx1000000);
	BinaryData->Latx1000000=(int32_t)round(Latx1000000);
	strncpy(BinaryData->Name,Name.c_str(),sizeof(BinaryData->Name));
	return true;
}

void EnrouteWaypoint::SetFromBinary(struct EnrouteWaypoint_Binary_struct *BinaryData) {
	ID.assign(BinaryData->ID,sizeof(BinaryData->ID));
	Lon=BinaryData->Lonx1000000/1000000.0;
	Lat=BinaryData->Latx1000000/1000000.0;
	Name.assign(BinaryData->Name,sizeof(BinaryData->Name));
	/* Reset the size to the actual number of characters */
	ID=ID.c_str();
	Name=Name.c_str();
}
