/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include "ImportSectionalsDVDVolumeChooserDialog.h"

#ifndef _GNU_SOURCE
/* Needed for strstr() in cstring */
#define _GNU_SOURCE
#endif

#ifndef _XOPEN_SOURCE
/* Needed for strptime() in time.h */
#define _XOPEN_SOURCE
#endif

#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <cstring>
#include <fstream>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <glibmm/i18n.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/main.h>
#include "Exception.h"
#include "Util.h"
#include "DataStore.h"

ImportSectionalsDVDVolumeChooserDialog::ImportSectionalsDVDVolumeChooserDialog(std::string *GladeFilename, std::string DataPath)
{
	try
	{
		refXml = Gnome::Glade::Xml::create(*GladeFilename, "ImportSectionalsDVDVolumeChooserDialog");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog(_("Error reading data file: ")+*GladeFilename+"\n"+_("Program cannot continue."),
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
	}

	refXml->get_widget("ImportSectionalsDVDVolumeChooserDialog",GtkDialog);
	refXml->get_widget("VolumeComboBox",VolumeComboBox);
	refXml->get_widget("EffectiveFromLabel",EffectiveFromLabel);
	refXml->get_widget("EffectiveToLabel",EffectiveToLabel);

	VolumeComboBox->signal_changed().connect(sigc::mem_fun(this,&ImportSectionalsDVDVolumeChooserDialog::on_VolumeComboBox_changed));

	/* Setup the Volume ComboBox */
	VolumeComborefTreeModel = Gtk::ListStore::create(VolumeComboColumns);
	VolumeComboBox->set_model(VolumeComborefTreeModel);

	this->DataPath=DataPath;
}

ImportSectionalsDVDVolumeChooserDialog::~ImportSectionalsDVDVolumeChooserDialog() {
	delete GtkDialog;
}

int ImportSectionalsDVDVolumeChooserDialog::Run() {
	PopulateVolumeComboBox();

	return GtkDialog->run();
}

void ImportSectionalsDVDVolumeChooserDialog::GetData(std::string *Volume, time_t *EffectiveFrom, time_t *EffectiveTo) {
	*Volume=this->Volume;
	*EffectiveFrom=this->EffectiveFrom;
	*EffectiveTo=this->EffectiveFrom+EffectiveDuration-1;
}

void ImportSectionalsDVDVolumeChooserDialog::on_VolumeComboBox_changed() {
	Gtk::TreeModel::iterator Activei;
	char buf[256];
	struct tm bdfrom, bdto;
	time_t EffectiveTo;

	Activei=VolumeComboBox->get_active();

	if (Activei) {		/* If anything is selected */
		Gtk::TreeModel::Row row = *Activei;
		Volume=row[VolumeComboColumns.Volume];
		EffectiveFrom=row[VolumeComboColumns.EffectiveFrom];
		EffectiveTo=EffectiveFrom+EffectiveDuration;

		try {
			Util::GmTime_r(&bdfrom,EffectiveFrom);
			Util::GmTime_r(&bdto,EffectiveTo);
		}
		catch (Exception e) {
			e.Show();
			return;
		}

		snprintf(buf,sizeof(buf),"%02u/%02u/%02u",bdfrom.tm_mon+1,bdfrom.tm_mday,bdfrom.tm_year-100);
		EffectiveFromLabel->set_text(buf);

		snprintf(buf,sizeof(buf),"%02u/%02u/%02u",bdto.tm_mon+1,bdto.tm_mday,bdto.tm_year-100);
		EffectiveToLabel->set_text(buf);
	} else {
		/* Nothing selected */
		EffectiveFrom=0;
		EffectiveFromLabel->set_text("");
		EffectiveToLabel->set_text("");
	}
}

void ImportSectionalsDVDVolumeChooserDialog::PopulateVolumeComboBox() {
	std::ifstream ifs;
	char buf[4096];
	std::string MinLine,MaxLine;
	time_t MetaFrom,MetaTo;
	struct tm bdfrom, bdto;
	int VolYear, VolIndex;
	DIR *dir;
	struct dirent *ent;

	/* Pick the first .html file to parse for data */
	dir=opendir(DataPath.c_str());
	if (!dir) {
		throw Exception("Error opening directory: "+DataPath+" ("+
			strerror(errno)+")",Exception::LEVEL_ERROR);
	}

	errno=0;
	do {
		ent=readdir(dir);
		if (ent && ent->d_name[0]!='.') {
			if (strstr(ent->d_name,".htm")!=NULL) {
				/* This is a match */
				/* ent->d_name is the name of our selected file */
				break;
			}
		}
	} while (ent);
	closedir(dir);
	if (errno || !ent || !ent->d_name) {
		throw Exception("Error selecting metadata file: "+DataPath+" ("+
			strerror(errno)+")",Exception::LEVEL_ERROR);
	}

	ifs.open((DataPath+"/"+ent->d_name).c_str());

	if (!ifs.good()) {
		throw Exception("Error opening metadata: "+DataPath+"/"+ent->d_name,Exception::LEVEL_ERROR);
	}

	do {
		ifs.getline(buf,sizeof(buf));
		if (strstr(buf,"dc.coverage.t.min")) {
			MinLine=buf;
		} else if (strstr(buf,"dc.coverage.t.max")) {
			MaxLine=buf;
		}
	} while (ifs.good() && (MinLine.length()==0 || MaxLine.length()==0));

	if (!ParseMetaDate(MinLine,&MetaFrom)) {
		throw Exception("Error parsing metadata",Exception::LEVEL_ERROR);
	}

	if (!ParseMetaDate(MaxLine,&MetaTo)) {
		throw Exception("Error parsing metadata",Exception::LEVEL_ERROR);
	}
	/* The "To" data needs to be adjusted forward 1 day minus 1 second to be consisted
	 * with the way the rest of the Effective dates are handled
	 */
	MetaTo=MetaTo+24*60*60-1;

	/* Scan through and set the possible volumes */
	VolumeComboBox->clear();

	try {
		Util::GmTime_r(&bdfrom,MetaFrom);
		Util::GmTime_r(&bdto,MetaTo);
	}
	catch (Exception e) {
		e.Show();
		return;
	}

	for (VolYear=bdfrom.tm_year-100;VolYear<=bdto.tm_year-100;VolYear++) {
		for (VolIndex=1;VolIndex<=14;VolIndex++) {
			std::ostringstream OVolStr;
			OVolStr<<std::setfill('0')<<std::setw(2)<<VolYear<<std::setw(2)<<VolIndex;
			if (
				MetaFrom<=Volume::ComputeEffectiveFrom(OVolStr.str()) &&
				MetaTo>=Volume::ComputeEffectiveTo(OVolStr.str())
			) {
				Gtk::TreeModel::Row row = *(VolumeComborefTreeModel->append());
				row[VolumeComboColumns.Volume] = OVolStr.str();
				row[VolumeComboColumns.EffectiveFrom] = Volume::ComputeEffectiveFrom(OVolStr.str());
			}
		}
	}

	VolumeComboBox->pack_start(VolumeComboColumns.Volume);
	VolumeComboBox->set_active(0);
}

bool ImportSectionalsDVDVolumeChooserDialog::ParseMetaDate(std::string Line, time_t *Date) {
	struct tm bdtime;

	if (Line.length()==0)
		return false;

	Line=Line.substr(Line.find("content="));		/* Strip off everything before 'content=' */
	Line=Line.substr(Line.find_first_of('\"')+1);	/* Strip off the leading 'content="' */
	Line=Line.erase(Line.find_first_of('\"'));		/* Strip off everything after the closing quote */
	/* Line is now in the form of 'FullMonthName, MDay, Year' or YYYYMMDD */

	memset(&bdtime,0,sizeof(bdtime));

	if (isalpha(Line[0])) {
		/* Line is FullMonthName, MDay, Year */
#ifndef WIN32
		strptime(Line.c_str(),"%B %d, %Y",&bdtime);
#else
		/* Line is similar to "April 9, 2009" */
		const char months[12][4]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};

		for (int i=0;i<12;i++) {
			if (Line.substr(0,3)==months[i])
				bdtime.tm_mon=i;
		}

		bdtime.tm_mday=atoi(Line.substr(Line.find_first_of(",")-2,2).c_str());
		bdtime.tm_year=atoi(Line.substr(Line.length()-4,4).c_str())-1900;
#endif
	} else {
		/* Line is YYYYMMDD */
		bdtime.tm_year=atoi(Line.substr(0,4).c_str())-1900;
		bdtime.tm_mon=atoi(Line.substr(4,2).c_str())-1;
		bdtime.tm_mday=atoi(Line.substr(6,2).c_str());
	}

	/* Things become effective at 0901Z */
	bdtime.tm_hour=9;
	bdtime.tm_min=1;

	*Date=Util::GMMkTime(&bdtime);
	return true;
}
