/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <cstring>
#include <sstream>
#include <fstream>
#include <iostream>
#include "Route.h"
#include "Geodesic.h"
#include "Exception.h"
#include "Util.h"

Route::Route()
{
}

void Route::Add(RoutePoint *RP)
{
	RoutePoints.push_back(RP);
	ClearLinRoutePoints();	/* These need to be recomputed */
}

void Route::Del(std::list<RoutePoint*>::iterator RPi)
{
	RoutePoints.erase(RPi);
	ClearLinRoutePoints();	/* These need to be recomputed */
}

void Route::Shift(std::list<RoutePoint*>::iterator RPi, int Direction)
{
	RoutePoint *TmpRP=*RPi;
	std::list<RoutePoint*>::iterator OtherRPi;

	if (Direction==1 && RPi!=RoutePoints.end()) {
		/* Swap the RoutePoint with the next one */
		OtherRPi=RPi;
		advance(OtherRPi,1);
		TmpRP=(*RPi);
		(*RPi)=(*OtherRPi);
		(*OtherRPi)=TmpRP;
	} else if (Direction==-1 && RPi!=RoutePoints.begin()){
		/* Swap the RoutePoint with the Previous one */
		OtherRPi=RPi;
		advance(OtherRPi,-1);
		TmpRP=(*RPi);
		(*RPi)=(*OtherRPi);
		(*OtherRPi)=TmpRP;
	}
	ClearLinRoutePoints();	/* These need to be recomputed */
}

void Route::Reverse()
{
	RoutePoints.reverse();
	ClearLinRoutePoints();	/* These need to be recomputed */
}

std::list<RoutePoint*> *Route::GetRoutePoints()
{
	return &RoutePoints;
}

/* Returns a linearized version of a great circle route */
std::list<RoutePoint*> *Route::GetLinearRoutePoints()
{
	if (RoutePoints.size()!=0 && LinRoutePoints.size()==0) {
		ComputeLinRoutePoints();
	}
	return &LinRoutePoints;
}

std::string Route::GetFilename() {
	return Filename;
}

/* Computes a linearized version of a great circle route */
void Route::ComputeLinRoutePoints()
{
	std::list<RoutePoint*>::iterator Pointi;

	/* Copy the exiting route to the new list as a staring point */
	ClearLinRoutePoints();
	for (Pointi=RoutePoints.begin();Pointi!=RoutePoints.end();Pointi++) {
		RoutePoint *RP=new RoutePoint(*Pointi);
		LinRoutePoints.push_back(RP);
	}

	if (LinRoutePoints.size()==1) {
		/* No need to linearize a single point */
		return;
	}

	std::list<RoutePoint*>::iterator PrevPointi;
	double Dist,Bear1,Bear2;
	for (Pointi=LinRoutePoints.begin();
		Pointi!=LinRoutePoints.end();Pointi++) {
		if (Pointi!=LinRoutePoints.begin()) {
			PrevPointi=Pointi;
			PrevPointi--;

			Geodesic::GreatCircleDistBear(
				(*PrevPointi)->GetLon(),(*PrevPointi)->GetLat(),
				(*Pointi)->GetLon(),(*Pointi)->GetLat(),
				&Dist,&Bear1,&Bear2);

			if (
				fabs(Bear2-Bear1)>0.5		/* The heading changes significantly */
				) {
				/* Subdivide this part of the route */
				RoutePoint *NewPoint = new RoutePoint();
				double NewLon,NewLat;
				/* Find the position on the great circle halfway between the
				 * previous and current point
				 */
				Geodesic::GreatCircleTravel(
					(*PrevPointi)->GetLon(),(*PrevPointi)->GetLat(),Dist/2.0,Bear1,
					&NewLon,&NewLat);
				/* Use that position and average the altitudes */
				NewPoint->SetLonLatAlt(NewLon,NewLat,
					((*PrevPointi)->GetAltitude()+(*Pointi)->GetAltitude())/2);
				LinRoutePoints.insert(Pointi,NewPoint);
				/* Continue checking from the previous point */
				Pointi=PrevPointi;
				Pointi--;
			}
		}
	}

	/* LinearRoutePoints is now complete */
	return;
}

bool Route::MakeBinary(struct Route_Binary_struct *BinaryData) {
	/* Sanity checking */
/*
	if (
		) {
		return false;
	}
*/

	return true;
}

void Route::SetFromBinary(struct Route_Binary_struct *BinaryData) {
}

bool Route::Load() {
	std::ifstream ifs;
	struct stat statdata;
	struct Route_Header_struct Header;
	uint32_t RunningCRC,HeaderCRC;
	uint32_t i;

	Util::ShellExpand(&Filename);

	if (stat(Filename.c_str(),&statdata)!=0) {
		throw Exception("Error processing "+Filename,Exception::LEVEL_ERROR);
	}

	ifs.open(Filename.c_str(),std::ifstream::binary);
	if (!ifs.good()) {
		throw Exception("Error opening "+Filename,Exception::LEVEL_ERROR);
	}

	/* Confirm the header is valid other than the CRC */
	ifs.read((char*)&Header,sizeof(Header));
	if (!ifs.good()) {
		throw Exception("File corrupt: cannod read header of "+Filename,Exception::LEVEL_ERROR);
	}
	if (strncmp(Header.Magic,ROUTE_FILEMAGIC,sizeof(ROUTE_FILEMAGIC)!=0)) {
		throw Exception("File corrupt: bad magic in "+Filename,Exception::LEVEL_ERROR);
	}
	if (Header.DataVersion!=ROUTE_HEADER_STRUCT_VER) {
		throw Exception("Unsupported data version in "+Filename,Exception::LEVEL_ERROR);
	}
	if (Header.FileSize!=(unsigned int)statdata.st_size) {
		throw Exception("File corrupt: size mismatch in "+Filename,Exception::LEVEL_ERROR);
	}
	/* Set the header filesize and CRC fields to zero and start computing the CRC */
	Header.FileSize=0;
	HeaderCRC=Header.CRC32;
	Header.CRC32=0;
	RunningCRC=CRC.Compute(&Header,sizeof(Header));

	/* Discard what's currently loaded */
	RoutePoints.clear();

	/* Load the data */
	try {
		for (i=0;i<Header.nRoutePoints;i++) {
			LoadRoutePoint(&RoutePoints,&ifs,&RunningCRC);
		}

		/* Check the CRC */
		if (HeaderCRC!=CRC.Finalize(RunningCRC)) {
			RoutePoints.clear();
			throw Exception("File corrupt - invalid CRC in "+Filename,Exception::LEVEL_ERROR);
		}
	} catch (Exception Ex) {
		ifs.close();
		RoutePoints.clear();
		throw Ex;
	}

	ifs.close();
	this->Filename=Filename;

	return true;
}

bool Route::Load(std::string Filename) {
	std::string OldFilename;

	OldFilename=Filename;
	this->Filename=Filename;

	try {
		Load();
	} catch (Exception Ex) {
		this->Filename=OldFilename;
		throw Ex;
	}

	return true;
}

void Route::Save() {
	std::ofstream ofs;
	struct Route_Header_struct Header;

	Util::ShellExpand(&Filename);
	ofs.open(Filename.c_str(),std::ofstream::binary);
	if (!ofs.good()) {
		throw Exception("Error opening "+Filename,Exception::LEVEL_ERROR);
	}

	memset(&Header,0,sizeof(Header));
	strncpy(Header.Magic,ROUTE_FILEMAGIC,sizeof(ROUTE_FILEMAGIC));
	Header.DataVersion=ROUTE_HEADER_STRUCT_VER;
	Header.nRoutePoints=RoutePoints.size();
	/* Leave FileSize set to 0 until after the CRC is computed */
	Header.CRC32=CRC.Compute(&Header,sizeof(Header));
	Header.FileSize+=sizeof(Header);

	ofs.seekp(sizeof(Header));	/* Skip the header for now */
	if (!ofs.good()) {
		ofs.close();
		throw Exception("Error processing "+Filename,Exception::LEVEL_ERROR);
	}

	SaveRoutePoints(&RoutePoints,&ofs,&Header.FileSize,&Header.CRC32);
	Header.CRC32=CRC.Finalize(Header.CRC32);
	ofs.seekp(0);	/* Write the header now that we have all the details */
	ofs.write((char*)&Header,sizeof(Header));
	if (!ofs.good()) {
		ofs.close();
		throw Exception("Error writing "+Filename,Exception::LEVEL_ERROR);
	}

	ofs.close();
}

void Route::Save(std::string Filename) {
	this->Filename=Filename;
	Save();
}

void Route::SaveRoutePoints(std::list<RoutePoint*> *RoutePointV, std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC) {
	std::list<RoutePoint*>::iterator RoutePointi;
	struct RoutePoint_Binary_struct Record;

	for (RoutePointi=RoutePointV->begin();RoutePointi!=RoutePointV->end();RoutePointi++) {
		memset(&Record,0,sizeof(Record));
		if (!((*RoutePointi)->MakeBinary(&Record))) {
			ofs->close();
			throw Exception("Error preparing route point data",Exception::LEVEL_ERROR);
		}
		*RunningCRC=CRC.Compute(&Record,sizeof(Record),*RunningCRC);
		ofs->write((char*)&Record,sizeof(Record));
		if (!ofs->good()) {
			ofs->close();
			throw Exception("Error writing route point data",Exception::LEVEL_ERROR);
		}
		*RunningSize+=sizeof(Record);
	}
}

void Route::LoadRoutePoint(std::list<RoutePoint*> *RoutePointV,std::ifstream *ifs, uint32_t *RunningCRC) {
	struct RoutePoint_Binary_struct Record;
	RoutePoint *RP;

	ifs->read((char*)&Record,sizeof(Record));
	if (!ifs->good()) {
		throw Exception("Error reading route point data",Exception::LEVEL_ERROR);
	}
	*RunningCRC=CRC.Compute(&Record,sizeof(Record),*RunningCRC);

	RP=new RoutePoint();
	if (!RP) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}
	RP->SetFromBinary(&Record);

	RoutePointV->push_back(RP);
}

void Route::Clear() {
	std::list<RoutePoint*>::iterator RoutePointi;

	for (RoutePointi=RoutePoints.begin();RoutePointi!=RoutePoints.end();++RoutePointi)
		delete (*RoutePointi);
	RoutePoints.clear();

	ClearLinRoutePoints();
}

void Route::ClearLinRoutePoints() {
	std::list<RoutePoint*>::iterator RoutePointi;

	for (RoutePointi=LinRoutePoints.begin();RoutePointi!=LinRoutePoints.end();++RoutePointi)
		delete (*RoutePointi);
	LinRoutePoints.clear();
}
